#!/usr/bin/env python

"""TKDO
A to-do list aggregator with due dates, repeating events, and a snooze button.

This pulls tasks from Vim-VimOutliner (.otl) files, sorts and filters them,
and outputs a single list with tasks prioritized in order.

Based vaguely on concepts from PyGTD.
"""

# Copyright (C) 2008 Scott Scriven
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# TODO: rename to something better...  maybe 'TKDO' or 'vo-todo'?
program_name = "tkdo"

cfg = None

import re

def counttabs(line):
    # FIXME: there has got to be a better way to do this
    i=0
    while i<len(line) and line[i] == '\t':
        i+=1
    return i

def datefmt(when=None):
    import time
    if not when:
        when = time.localtime()
    if isinstance(when, float):
        when = time.localtime(when)
    text = time.strftime(cfg.datefmt, when)
    return text

def dateparse(text):
    import time
    date = time.strptime(text, cfg.datefmt)
    #date = time.mktime(date)

    #print "dateparse(%s): %s" % (text, repr(date))
    return date

def date_add(date, num, unit):
    import time
    units = dict(h=(1,3), d=(1,2), w=(7,2), m=(1,1), y=(1,0))
    if unit not in units:
        raise ValueError, "bad unit: %s" % (unit)
    date = list(date)
    amt, field = units[unit]
    amt *= num
    date[field] += int(round(amt))
    date = tuple(date)
    # convert to epoch and back, to get a valid tuple
    date = time.mktime(date)
    date = time.localtime(date)

    return date

def date_diff(a, b):
    import time
    at = time.mktime(a)
    bt = time.mktime(b)
    diff = at - bt
    return diff

def duration(num, unit):
    """convert a readable time length into number of seconds.
    units are: h d w m y
    Use date_add instead, if you care about issues like leap years
    or daylight savings time.  This code gets those WRONG.
    """
    units = dict(
            h=60.0 * 60,
            d=60.0 * 60 * 24,
            w=60.0 * 60 * 24 * 7,
            m=60.0 * 60 * 24 * (365.0/12),
            y=60.0 * 60 * 24 * 365,
            )
    return num * units[unit]

def iround(num):
    "convert num (float) to str, and remove trailing zeroes"
    text = str(num)
    if text[-2:] == ".0":
        text = text[:-2]
    return text

def open_url(url, mode="rb"):
    """abstract file access to make URLs and remote files work"""
    # TODO: detect/save mtime (?)
    # (could be a pain for remote files)
    def expand_home(url):
        import os
        path = url
        path = path.replace("~", os.environ["HOME"])
        path = path.replace("$HOME", os.environ["HOME"])
        return path

    if "://" not in url:
        url = "file:///" + url
    # protocol:// (?), hostname (?), :port (?), path
    pat = re.compile(r"(\w+://)?([^/:]+)?(:\d+)?/(.+)")
    found = pat.search(url)
    if not found:
        raise ValueError, "invalid URL: %s" % repr(url)

    protocol = found.group(1)[:-3]
    hostname = found.group(2)
    portnum = found.group(3)
    if portnum: portnum = portnum[1:]
    path = found.group(4)

    if protocol == "file":
        path = expand_home(path)
        fp = open(path, mode)
        return fp
    # TODO: add protocol for ssh, scp, sftp, or rsync
    else:
        raise ValueError, "unsupported protocol %s" % repr(protocol)

class Task:
    # removed 'project' concept, since I didn't use it
    # TODO: add 'next' type, which only shows the first child (ala GTD)
    valid_types = ('task', 'idea', 'ignored')
    active_types = ('task', 'idea')

    meta_pats = dict(
            type = re.compile(r"T=(\S+)"),
            # inferred from due_date
            # TODO: make HH:MM optional
            # TODO: add non-recurring due dates
            due_date = re.compile(r"D=(\d{4}-\d\d-\d\d@\d\d:\d\d)?\+([\.\d]+)([hdwmy])"),
            lead_time = re.compile(r"L=([\.\d]+)([hHdDwWmMyY])"),
            snooze_time = re.compile(r"Z=(\d{4}-\d\d-\d\d@\d\d:\d\d)\+([\.\d]+)([hdwmy])"),
            last_done = re.compile(r"X=(\d)?,?(\d{4}-\d\d-\d\d@\d\d:\d\d)"),
            importance = re.compile(r"I=([\d]+)"),
            depth = re.compile(r"depth=([\d]+)"),
            )

    def __init__(self, parent=None, lines=[], lineno=0):
        self.parent = parent
        self.lines = lines
        self.lineno = lineno

        self._reset()

        self._parse_meta()

        # create child tasks, if we're not too deep already
        depth = self.inherit('depth')
        if (self.type in self.active_types) and \
                (depth > self.level + 1): # depth is 1-based, level is 0-based
            self.load_children()

    def _reset(self):
        self.level = 0
        if self.parent:
            self.level = self.parent.level + 1

        self.tasks = []

        self.alias = None
        self.title = None
        self.active = None
        self.type = None
        self.done = None
        self.last_done = None
        self.was_done = None
        self.due_date = None
        self.recur_type = None
        self.lead_time = None
        self.snoozed = None
        self.contexts = []
        self.importance = None
        self.depth = None

    def get_root(self):
        "find and return our root node"
        t = self
        while t.parent:
            t = t.parent
        return t

    def load_children(self):
        indent = '\t' * (self.level + 1)
        # find next-level items
        # must be indented enough, but not too much
        # must not be our meta line
        firstlines = [num for num,line in enumerate(self.lines)
                if (line.startswith(indent))
                and (not line[len(indent):].startswith('\t'))
                and (not line[len(indent):].startswith('; TKDO: '))]

        child_tasks = []
        for tnum in range(len(firstlines)):
            start = firstlines[tnum]
            try:
                next = firstlines[tnum+1]
            except:
                next = None
            tlines = self.lines[start:next]
            t = Task(parent=self, lines=tlines, lineno=start+self.lineno)
            child_tasks.append(t)

        self.tasks = child_tasks

    def has_active_subtasks(self):
        for t in self.tasks:
            if t.type in self.active_types:
                return True
        return False

    def full_title(self):
        parts = []
        parent = self
        while parent:
            if parent.alias: text = parent.alias
            else: text = parent.title
            parts.insert(0, text)
            parent = parent.parent
        return " :: ".join(parts)

    def inherit(self, key):
        val = getattr(self, key)

        parent = self.parent
        while parent and (val is None):
            val = getattr(parent, key)
            parent = parent.parent
        return val

    def inherit_list(self, key):
        val = getattr(self, key)

        # collect values
        parent = self.parent
        while parent:
            pval = getattr(parent, val)
            val.extend(pval)
            parent = parent.parent

        # remove duplicates
        dv = dict(zip(val, [1]*len(val)))
        val = dv.keys()

        return val

    def _parse_meta(self):
        if len(self.lines) >= 2:
            tline, mline = self.lines[:2]
        elif len(self.lines) >= 1:
            tline = self.lines[0]
            mline = None
        else:
            tline = mline = None
        self._parse_firstline(tline)
        self._parse_metaline(mline)

    def _parse_firstline(self, line):
        if not line:
            self.type = 'ignored'
            return

        # amount of indentation; depth in outline tree
        #self.level = counttabs(line)

        # remove extra stuff from title
        mangle = (
                ('[_] ', ''),
                ('[X] ', ''),
                ('[Z] ', ''), # snoozed
                (' <--', ''),
                )
        self.title = line.strip()
        for search, replace in mangle:
            self.title = self.title.replace(search, replace)

        # leave None instead of False, so inheritance will work
        if " <--" in line: self.active = True

        # ? also, load extra data like contexts, last date completed, etc
        if ('[_] ' in line) or ('[Z] ' in line):
            self.type = 'task'
            self.done = False
        elif ('[X] ' in line):
            self.type = 'task'
            self.done = True
        else:
            self.type = 'ignored'
            self.title = line
            self.done = False

    def _parse_metaline(self, line):
        """read the line containing all the task attributes,
        assigning any discovered values to self.
        example:
            ; TKDO: T=task D=2008-04-14@17:00+1m L=1w Z=2008-04-00@12:12+1w X=1,2008-03-14@17:33 @work @online
        """
        if not line:
            return

        ## ensure line is indented once more than self
        ## (turns out it's better to let the caller ensure this is the right line)
        #if counttabs(line) != counttabs(self.lines[0]) + 1:
        #    return
        if not line.strip().startswith('; TKDO: '):
            return

        #print line

        result = {}

        # contexts
        parts = line.split()
        contexts = [x[1:] for x in parts if x.startswith('@')]
        result['contexts'] = contexts
        self.contexts = contexts

        # grab what I can via regexes
        for key, pattern in self.meta_pats.iteritems():
            found = pattern.search(line)
            if found:
                value = found.groups()
                result[key] = value

        #for key, values in result.iteritems():
        #    print "%s: %s" % (key, values)

        if 'depth' in result:
            x = result['depth'][0]
            self.depth = int(x)

        if 'type' in result:
            t = result['type'][0]
            if t in self.valid_types:
                self.type = t
            if self.type == 'idea':
                self.importance = self.inherit('importance') / 2.0

        if 'importance' in result:
            x = result['importance'][0]
            x = float(x) / 100.0
            if self.type == 'idea':
                x /= 2
            self.importance = x

        if 'last_done' in result:
            x = result['last_done']
            self.was_done = self.done
            if x[0] is not None:
                self.was_done = not not int(x[0])
            self.last_done = dateparse(x[1])

            # user checked off this item since last time we looked
            #if self.done and not self.was_done:
            #    self.last_done = time.localtime()

        if 'lead_time' in result:
            r = result['lead_time']
            self.lead_time = (float(r[0]), r[1].lower())

        if 'snooze_time' in result:
            x = result['snooze_time']
            base = dateparse(x[0])
            self.snoozed = (base, float(x[1]), x[2].lower())

        if 'due_date' in result:
            r = result['due_date']
            base = None
            self.recur_type = 'float'
            if r[0]:
                base = dateparse(r[0])
                self.recur_type = 'fixed'
            if r[1]:
                self.due_date = (base, float(r[1]), r[2].lower())
            else:
                self.due_date = (base, None, None)

        """
        if self.recur_type == 'fixed':
            # TODO: magic with self.done and adjusting our due date
            # if not done:
            #   no change
            # elif done:
            #   while last completed > due date - lead time:
            #     bump due date forward one unit
            #   if last completed < due date - lead time:
            #     no change
            #   else:
            #     done = 0
        """

    def complete(self):
        import time
        self.done = 1
        self.last_done = time.localtime()

        if self.recur_type:
            # update our due date
            due_date = self.due_date
            last_done = time.mktime(self.last_done)
            lead = 0.0 # TODO: unsure if this is correct
            interval = due_date[1:3]
            if self.recur_type == 'float':
                # no explicit due date
                pass
            elif self.recur_type == 'fixed':
                due = due_date[0]
                while (time.mktime(due) - lead) < last_done:
                    due = date_add(due, *interval)

                foo = list(due_date)
                foo[0] = due
                self.due_date = tuple(foo)

    def snooze(self, howlong):
        """Mark this item as snoozed, starting now, for 'howlong' time.
        'howlong' should be a number and a time unit [hdwmy]
        """
        pat = re.compile(r"^\+?([\.\d]*)\s*([hHdDwWmMyY])$")

        found = pat.search(howlong)
        if not found:
            raise ValueError, "Invalid snooze length."

        num = found.group(1)
        if num:
            num = float(num)
        else:
            num = 1.0
        unit = found.group(2).lower()

        import time
        now = time.localtime()
        self.snoozed = (now, num, unit)

    def set_active(self, state=None):
        if state is None:
            # toggle
            if self.active:
                self.active = None
            else:
                self.active = True
        else: # set to True or None (not False)
            if state:
                self.active = True
            else:
                self.active = None

    def make_metaline(self):
        "generate a meta line"
        data = []

        # somewhat ignores same-as-parent attributes while saving

        # type
        if self.type != 'task':
            data.append('T=%s' % (self.type))

        # depth
        if self.depth is not None:
            data.append('depth=%s' % (self.depth))

        # importance
        # (only save if non-default, and different than parent)
        imp = self.inherit('importance')
        pimp = None
        if self.parent:
            pimp = self.parent.inherit('importance')
        if self.type == 'idea': imp *= 2
        if (self.importance is not None)  \
                and  (pimp != imp) \
                and  ((imp < 0.499) or (imp > 0.501)):
            data.append("I=%s" % (int(imp*100)))

        # due date / recurrance info
        if self.due_date:
            x = self.due_date
            base = ''
            if x[0]:
                base = datefmt(x[0])
            if x[1]:
                data.append("D=%s+%s%s" % (base, iround(x[1]), x[2]))
            else:
                data.append("D=%s" % (base))

        # lead time
        if self.lead_time:
            x = self.lead_time
            data.append("L=%s%s" % (iround(x[0]), x[1]))

        # last completion date
        if self.done and not self.last_done:
            data.append('X=%s,%s' % (int(self.done),datefmt()))
        elif self.last_done:
            data.append('X=%s,%s' % (int(self.done),datefmt(self.last_done)))

        # snooze data
        if self.snoozed:
            # TODO: omit, if snooze end time has passed
            x = self.snoozed
            base = datefmt(x[0])
            data.append("Z=%s+%s%s" % (base, iround(x[1]), x[2]))

        # contexts
        for c in self.contexts:
            data.append('@%s' % (c))

        line = ('\t' * (self.level+1)) + '; TKDO: ' + " ".join(data)
        # avoid meta lines with nothing but an X tag
        # (those just don't matter)
        if (len(data) == 1) and (data[0].startswith('X=')):
            return None

        elif data:
            return line

        return None

    def to_otl(self):
        """Return text representing self in vimoutliner .otl format.
        """
        #print "Task.to_otl(): %s" % (self.title)
        # ignore same-as-parent attributes while saving
        if self.type == 'ignored':
            text = "\n".join(self.lines)
            return text

        import time

        lines = self.lines[:]
        depth = self.level
        # build a main line
        line = self.title
        active = self.active
        if self.type in self.active_types:
            if self.is_snoozed():
                if self.done and (not self.recur_type):
                    line = '[X] ' + line
                else:
                    line = '[Z] ' + line
            elif self.recur_type:
                if self.done or (self._score < 0.001):
                    line = '[Z] ' + line
                else:
                    line = '[_] ' + line
            elif self.done:
                line = '[X] ' + line
                active = False # completed one-time tasks are never active
            else:
                line = '[_] ' + line
        if active: line = line + " <--"
        line = ('\t' * depth) + line
        lines[0] = line

        # add a meta line
        line = self.make_metaline()
        if line:
            lines.insert(1, line)
            lead = lines[:2]
            remaining = lines[2:]
        else:
            lead = lines[:1]
            remaining = lines[1:]

        # recurse into child tasks when depth allows
        if self.tasks:
            lines = lead
            for t in self.tasks:
                lines.append(t.to_otl())

        else:
            # remove old/extra meta lines
            lines = lead + [l for l in remaining
                    if not l.strip().startswith('; TKDO: ')
                    or (counttabs(l) != depth+1) ]

        return "\n".join(lines)

    def is_snoozed(self):
        import time
        now = time.time()

        snoozed = self.inherit('snoozed')
        if snoozed:
            snoozed_at = snoozed[0]
            interval = snoozed[1:3]
            due = date_add(snoozed_at, *interval)
            due = time.mktime(due)
            waketime = due - (duration(*interval) / 2.0)
            if waketime > now: # not time to wake up
                return True

            if now > due: # snooze has passed entirely
                self.snoozed = None

        return False

    def timeliness(self, when=None):
        """Compute the urgency, or scheduling pressure for this task.
        Returns a float from 0.0 to a maximum urgency (normally 1.5).
        """
        # FIXME: should probably eliminate side effects
        import time
        t = 1.0
        if when:
            now = when
        else:
            now = time.time()

        # snooze hides an item for snooze_length / 2,
        # then lets it bubble up for the other half of the snooze_length.
        parent = self
        while parent: # snooze from all parents, not just the first match
            snoozed = parent.snoozed
            if snoozed:
                start = time.mktime(snoozed[0])
                # due = start + length
                due = date_add(snoozed[0], *snoozed[1:3])
                due = time.mktime(due)
                length = due - start
                window = length/2
                until = due - now
                urgency = 1.0 - (until / window)
                urgency = min(1.0, max(0.0, urgency)) # clip to range (0, 1)
                t *= urgency

            parent = parent.parent

        # regular recurring tasks
        due_date = self.inherit('due_date')
        if due_date is not None:
            interval = due_date[1:3]

            # figure out how long our ramp-up window is
            lead_time = self.inherit('lead_time')
            if lead_time:
                lead = duration(*lead_time)
            else:
                lead = duration(*interval) / 2.0

            # figure out when we were last completed
            last_done = self.inherit('last_done')
            done = self.inherit('done')
            if last_done: # has history
                last_done = time.mktime(last_done)
            elif done: # completed, but no history attached
                # assume it was completed just now
                # FIXME: write to parent here instead?
                self.last_done = time.localtime(now)
                last_done = time.mktime(self.last_done)
            else: # has never been completed
                last_done = 0.0

            # handle floating recurring events
            recur_type = self.inherit('recur_type')
            if recur_type == 'float':
                due = last_done + duration(*interval)
                start = due - lead
                if start > now: # hide if it hasn't reached the lead-in time
                    t = 0.0
                else: # (start < now < due) or (past due)
                    until = due - now
                    urgency = 1.0 - (until / lead)
                    urgency = min(1.25, max(0.0, urgency)) # clip to range (0, 1.25)
                    t *= urgency
                    #print "urgency: %s" % (urgency)

            # handle fixed recurring events
            elif recur_type == 'fixed':
                # base = due date = original recur date
                # while due date < date last completed:
                #     due date = due date + recur interval
                # ? recur date = due date
                due = due_date[0]
                while (time.mktime(due) - lead) < last_done:
                    due = date_add(due, *interval)

                foo = list(due_date)
                foo[0] = due
                # FIXME: write to parent here instead?
                self.due_date = tuple(foo)

                due = time.mktime(due)
                start = due - lead

                if start > now: # hide if it hasn't reached the lead-in time
                    t = 0.0
                else: # start < now < due :
                    until = due - now
                    urgency = 1.0 - (until / lead)
                    urgency = min(1.5, max(0.0, urgency)) # clip to range (0, 1.5)
                    t *= urgency
                    #print "urgency: %s" % (urgency)

        #print "timeliness: %s" % (t)
        return t

    def score(self, when=None):
        # plain old tasks with no special attributes are omitted if completed
        if self.done and (self.type in self.active_types) \
                and not self.due_date:
                    self._score = 0.0
        else:
            self._score = self.inherit('importance') * self.timeliness(when)
            ## quantize, to avoid the reverse order effect caused by time 
            ## passing while scoring items
            #self._score = round(self._score, 2)
        if self.inherit('active'):
            self._score *= 1.25
        return self._score

class Taskfile(Task):
    def __init__(self, title, url, **kwargs):
        self.title = title
        self.url = url
        Task.__init__(self)
        self._reset()

    def _reset(self):
        tmp = self.title, self.url
        Task._reset(self)
        self.title, self.url = tmp

        self.parent = None # should always be None
        self.type = 'task'
        self.level = -1 # 0 is a non-indented task; we're one up from that
        self.depth = 1 # top-level tasks only
        self._changed = 0 # track whether we need to save
        if self.importance is None:
            self.importance = 0.5

    def __repr__(self):
        text = "Taskfile(%s, %s)" % (repr(self.title), repr(self.url))
        return text

    def to_otl(self):
        text = []
        # delete our meta line
        tasks = self.root_tasks
        try:
            assert(tasks[0].lines[0].startswith('; TKDO: '))
            del tasks[0]
        except:
            pass

        # make a new meta line
        # TODO: put metaline at end of file
        line = self.make_metaline()
        if line:
            text.append(line.strip())

        # add the rest 
        for t in tasks:
            text.append(t.to_otl())

        # remove trailing blank lines
        while text[-1:] == ['']:
            del text[-1]
        # ensure file ends with '\n'
        if text  and  (not text[-1].endswith('\n')):
            text.append('')

        text = "\n".join(text)
        return text

    def load(self):
        self._reset()

        fp = open_url(self.url)
        text = fp.read()
        fp.close()
        self.lines = text.split("\n")

        self.tasks = []

        # parse taskfile attributes from first line of file
        # TODO: grab meta line from either beginning or end of file
        if (len(self.lines) > 0) and self.lines[0].startswith('; TKDO: '):
            line = self.lines[0]
            self._parse_metaline(line)

        if (self.type == 'ignored') or (self.depth < 1):
            # don't try to load tasks
            return

        # find top-level items
        toplines = [num for num,line in enumerate(self.lines) if (not line.startswith('\t'))]

        root_tasks = []
        for tnum in range(len(toplines)):
            start = toplines[tnum]
            try:
                next = toplines[tnum+1]
            except:
                next = None
            tlines = self.lines[start:next]
            t = Task(parent=self, lines=tlines, lineno=start+1)
            root_tasks.append(t)
        self.root_tasks = root_tasks

        leaves = []
        def add_leaves(mytasks):
            for t in mytasks:
                if not t.has_active_subtasks():
                    leaves.append(t)
                else:
                    add_leaves(t.tasks)
        add_leaves(root_tasks)
        self.leaves = leaves
        self.tasks = self.leaves

    def save(self):
        # TODO: move old file to .bak first
        # FIXME: recurring tasks should be marked done or not before saving
        fp = open_url(self.url, "wb")
        text = self.to_otl()
        fp.write(text)
        fp.close()

class Masterlist:
    def __init__(self, taskfiles=None):
        self.taskfiles = []
        if taskfiles:
            self.taskfiles = taskfiles[:]
        self.tasks = []

    def load(self):
        for tf in self.taskfiles:
            #print repr(tf)
            tf.load()
            #print "  %s lines" % (len(tf.lines))
        self.update()

    def update(self):
        self.tasks = []
        for tf in self.taskfiles:
            tasks = [t for t in tf.tasks if t.type in t.active_types]
            self.tasks.extend(tasks)
            unfinished = [t for t in tasks if not t.done]
            finished = [t for t in tasks if t.done]
            active = [t for t in tasks if t.active]
            #print "  %s tasks (%s/%s active, %s done)" % (len(tasks), len(active), 
            #        len(unfinished), len(finished))

            #for t in tasks:
            ##    print "  %s %s: %s" % (int(t.score() * 100), t.parent.title, t.title)
            #    print t.to_otl()

    def sort(self):
        # freeze time while scoring...
        # otherwise, same-score scheduled items end up in reverse order
        # (because time passed between samples)
        import time
        now = time.time()
        tasks = [(t.score(now), -t.lineno, t.title, t) for t in self.tasks]
        tasks.sort()
        tasks.reverse()
        tasks = [t[-1] for t in tasks]
        self.tasks = tasks

    def print_list(self):
        unfinished = [t for t in self.tasks if not t.done]
        finished = [t for t in self.tasks if t.done]
        active = [t for t in self.tasks if t.active]
        print "%s total tasks (%s/%s active, %s done)" % (len(self.tasks), 
                len(active), len(unfinished), len(finished))

        for t in self.tasks:
            if t._score > 0.001:
                active = ""
                if t.active: active = " <--"
                print "  %3s %s%s" % (int(t.score() * 100), t.full_title(), active)
            #print t.to_otl()

class MasterlistTUI(Masterlist):
    def __init__(self, *args, **kwargs):
        Masterlist.__init__(self, *args, **kwargs)

class MasterlistGUI(Masterlist):
    def __init__(self, *args, **kwargs):
        Masterlist.__init__(self, *args, **kwargs)
        import pygtk
        pygtk.require('2.0')
        import gtk
        self.gtk = gtk
        self.filters = []

    def run(self):
        self._build_ui()
        self.update_treestore()
        try:
            self.gtk.main()
        except KeyboardInterrupt:
            pass

    def update_treestore(self, *args):
        path, column = self.treeview.get_cursor()
        self.treestore.clear()
        # TODO: move the filter stuff to Masterlist
        num_visible = 0
        num_total = len(self.tasks)
        num_done = len([t for t in self.tasks if t.done and (not t.recur_type)])
        # show/hide zero-score items
        # show/hide done items
        for t in self.tasks:
            if cfg.hide_done  and  t.done  and  (t._score < 0.001):
                continue
            if cfg.hide_zero  and  (t._score < 0.001)  \
                    and  (t.inherit('snoozed')  or  t.inherit('recur_type')):
                continue

            done = ""
            if t.done: done = "[X] "
            active = ""
            if t.active: active = " <--"
            title = "%s%s%s" % (done, t.full_title(), active)

            skipped = False
            for search in self.filters:
                if search.lower() not in title.lower():
                    skipped = True

            if not skipped:
                num_visible += 1
                data = [t, int(t.score()*100), title]
                branch = self.treestore.append(None, data)
                if len(t.lines) > 1:
                    data = ['', '', "\n".join(t.lines[1:])]
                    self.treestore.append(branch, data)
        if path:
            self.treeview.set_cursor(path)

        # set window title to indicate number of visible tasks
        self.window.set_title("%s: %s/%s tasks (%s done) in %s files" % 
                (program_name, num_visible, num_total, num_done, 
                    len(self.taskfiles)))

    def _build_ui(self):
        g = self.gtk
        self.window = g.Window(g.WINDOW_TOPLEVEL)
        self.window.set_title(program_name)
        self.window.connect("destroy", self._quit)
        self.window.connect("delete_event", self._quit)
        # TODO: sane window sizing
        self.window.set_size_request(500,600)
        vb = g.VBox(homogeneous=False, spacing=0)
        self.window.add(vb)

        menus = self._build_menus()
        vb.pack_start(menus, False, True, 0)

        hb = g.HBox(homogeneous=False, spacing=0)
        vb.pack_start(hb)

        self.treestore = g.TreeStore(object, str, str)
        self.treeview = g.TreeView(self.treestore)
        self.treeview.set_enable_search(False)
        self.treeview.connect('key-press-event', self._tree_keypress)
        # FIXME: scrollbar can be invisible, off the right side
        # FIXME: scroll wheel doesn't work
        # FIXME: scrollbar use doesn't keep cursor onscreen
        hb.pack_start(self.treeview)
        va = self.treeview.get_vadjustment()
        scroll = g.VScrollbar(va)
        hb.pack_start(scroll, expand=False, fill=False)

        colnum = 0

        # just the widget to expand the branch
        self.expand_column = g.TreeViewColumn('')
        self.treeview.append_column(self.expand_column)
        colnum += 1

        # task score
        self.score_column = g.TreeViewColumn('score')
        self.treeview.append_column(self.score_column)
        cell = g.CellRendererText()
        self.score_column.pack_start(cell, True)
        self.score_column.add_attribute(cell, 'text', colnum)
        colnum += 1

        # task title
        self.title_column = g.TreeViewColumn('title')
        self.treeview.append_column(self.title_column)
        cell = g.CellRendererText()
        self.title_column.pack_start(cell, True)
        self.title_column.add_attribute(cell, 'text', colnum)
        colnum += 1

        # TODO: status bar
        hb = g.HBox()
        vb.pack_start(hb, expand=False, fill=False)

        self._status = g.Label()
        self._status.set_justify(g.JUSTIFY_LEFT)
        self._status.set_alignment(0.0, 0.5)
        self._status.set_padding(5,5)
        hb.pack_start(self._status)

        # entry box for search, snooze, vim command emulation, etc
        self._status_entry = g.Entry()
        self._status_entry.connect('activate', self._activate_status_entry)
        hb.pack_start(self._status_entry)

        self.window.show_all()
        self._status_entry.hide()
        self.treeview.grab_focus()

    def _build_menus(self):
        "create some menu widgets"
        # FIXME: DeprecationWarning: use gtk.UIManager instead
        menu_items = (
            ('/_File', None, None, 0, '<Branch>'),
            ("/File/_Save current task's file", '<control>S', self._save_task_cb, 0, None),
            # TODO: save config file
            ('/File/Save config file', None, None, 0, None),
            ('/File/_Refresh', '<control>R', self._refresh, 0, None),
            ('/File/Re_load All', None, self._reload_all, 0, None),
            ('/File/_Quit', '<control>Q', self._quit, 0, None),
            ('/_Edit', None, None, 0, '<Branch>'),
            ('/Edit/_Go to (edit) taskfile', '<control>G', self._gvim_task, 0, None),
            ('/_View', None, None, 0, '<Branch>'),
            ('/View/Expand or Collapse Task', None, self._toggle_expand, 0, None),
            ('/View/_Filter list', '<control>F', self._start_search, 0, None),
            # TODO: add UI to show/hide zero-score items
            ('/View/Hide\/Show snoozed tasks', None, None, 0, None),
            # TODO: add UI to show/hide done items
            ('/View/Hide\/Show done tasks', None, None, 0, None),
            ('/_Actions', None, None, 0, '<Branch>'),
            ('/Actions/Snoo_ze item', '<control>Z', self._start_snooze,0, None),
            ('/Actions/Mark item as done', '<control>X', self._complete_item, 0, None),
            ('/Actions/Toggle _Active', None, self._toggle_active, 0, None),
            # TODO: set due date (needs a date/duration chooser)
            # TODO: add a dialog listing all keybindings
            #       (by key and by function)
            )
        g = self.gtk
        accel_group = g.AccelGroup()
        factory = g.ItemFactory(g.MenuBar, '<main>', accel_group)
        factory.create_items(menu_items)
        self.window.add_accel_group(accel_group)
        self.menu_factory = factory
        menubar = factory.get_widget('<main>')
        return menubar

    def _tree_keypress(self, widget, event):
        keyval = event.keyval
        name = self.gtk.gdk.keyval_name(keyval)
        string = event.string
        #print "key: %s, %s, %s" % (keyval, name, string)
        # TODO: convert keys/responses to a table lookup
        # TODO: make key bindings configurable
        if name == "g":
            self._gvim_task()
        # start a search
        elif name in ('slash', 'L'): # "L" is what I use in mutt
            self._start_search()
        # cancel search/filter with escape
        elif name == 'Escape':
            if self.filters:
                self.filters = []
                self._status_entry_mode = 'search'
                self._activate_search(None)
        elif name == 'space':
            self._toggle_expand()
        elif name == 'x':
            self._complete_item()
        elif name == 'z':
            self._start_snooze()
        elif name == 'colon':
            self._start_vicommand()
        elif name == 'a':
            self._toggle_active()
        elif name == 't': # TODO: set task type
            #self._set_type()
            pass
        elif name == 'd': # TODO: set due date
            #self._set_due_date()
            pass
        elif name == 'k': # TODO: mark as [S] skipped
            #self._set_due_date()
            pass

    def _get_current_task(self, *args):
        path, column = self.treeview.get_cursor()
        if not path:
            self.set_status('Select an item first.')
            return None
        data = self.treestore[path]
        t = data[0]
        if not t:
            self.set_status('Selected item is not a task.')
            return None
        return t

    def _save_task_cb(self, *args):
        t = self._get_current_task()
        if not t:
            return
        tf = t.get_root()
        self._save_task(t, "Saved taskfile: %s" % (tf.title))

    def _save_task(self, t, msg):
        # find task file
        tf = t.get_root()
        # save
        self.set_status("Saving %s" % (tf.url))
        tf.save()
        self.set_status("Reloading %s" % (tf.url))
        tf.load()
        self._refresh()
        self.set_status(msg)

    def gtk_flush(self, *args):
        while self.gtk.events_pending():
            self.gtk.main_iteration()

    def _refresh(self, *args):
        self.update()
        self.sort()
        self.update_treestore()

    def _reload_all(self, *args):
        for tf in self.taskfiles:
            self.set_status("Loading %s" % (tf.url))
            tf.load()
        self._refresh()
        self.set_status("Reloaded all task files.")

    def set_status(self, text):
        self._status.set_text(text)
        self.gtk_flush()

    def _gvim_task(self, *args):
        t = self._get_current_task()
        if not t:
            return

        tf = t.get_root()
        self.edit_taskfile(tf, t)
        tf.load()
        self._refresh()

    def _complete_item(self, *args):
        t = self._get_current_task()
        if not t:
            return

        # TODO: show a brief strike-thru
        import time

        # mark done
        t.complete()

        # save
        self._save_task(t, "Completed: %s" % (t.full_title()))

    def _toggle_active(self, *args):
        t = self._get_current_task()
        if not t:
            return

        # toggle active
        t.set_active()

        self._save_task(t, "Toggled Active: %s" % (t.full_title()))

    def edit_taskfile(self, tf, t):
        url = tf.url
        lineno = t.lineno
        # FIXME: escape command parameters
        # FIXME: make local copy, edit, save via url abstraction
        cmd = "gvim -f +%s %s" % (lineno, str(url))
        print cmd
        import os
        self.set_status(cmd)
        os.system(cmd)
        self.set_status("Done: " + cmd)

    def _toggle_expand(self, *args):
        path, column = self.treeview.get_cursor()
        if not path:
            return
        if self.treeview.row_expanded(path):
            self.treeview.collapse_row(path)
        else:
            self.treeview.expand_row(path, open_all=False)

    def _expand_row(self, *args):
        path, column = self.treeview.get_cursor()
        if not path:
            return
        self.treeview.expand_row(path, open_all=False)

    def _collapse_row(self, *args):
        path, column = self.treeview.get_cursor()
        if not path:
            return
        self.treeview.collapse_row(path)

    def _start_entry(self, mode, text, needs_task=False):
        "open an entry box, get some text, and pass it to a handler"
        # save the current task for later, in case the handler wants it
        t = self._get_current_task()
        self._status_entry_task = t
        if needs_task and not t:
            return

        # get user input
        self.set_status(text)
        self._status_entry_mode = mode
        self._status_entry.set_text('')
        self._status_entry.show()
        self._status_entry.grab_focus()

    def _start_search(self, *args):
        self._start_entry('search', "Search:")

    def _start_snooze(self, *args):
        self._start_entry('snooze', "Snooze: (t|p+|f)?[0-9]*[hdwmy]", needs_task=True)

    def _start_vicommand(self, *args):
        self._start_entry('vicommand', "Vi emulation:")

    def _activate_status_entry(self, *args):
        self._status_entry.hide()
        text = self._status_entry.get_text()
        self._status_entry.set_text('')

        # find a handler, and call it
        mode = self._status_entry_mode
        try:
            func = getattr(self, "_activate_%s" % (mode))
        except:
            raise ValueError, "invalid mode %s" % (mode)

        func(text)

        self._status_entry_mode = None
        self.treeview.grab_focus()

    def _activate_search(self, text):
        # filter visible data
        if text:
            self.filters.append(text)
            self.set_status("Search done.")
        else:
            self.filters = []
            self.set_status("Filter removed.")
        self.update_treestore()
        self._refresh()

    def _activate_snooze(self, text):
        t = self._status_entry_task
        self._status_entry_task = None
        if not t:
            self.set_status('Select a task first.')
            return

        # TODO: move this logic into Task.snooze() ?
        # (makes it more difficult to set the status correctly though)
        if text.startswith('t'): # snooze task (default)
            text = text[1:]
        elif text.startswith('p'): # snooze parent (or ancestor)
            while text.startswith('p'):
                text = text[1:]
                if t.parent:
                    t = t.parent
        elif text.startswith('f'): # snooze entire file
            text = text[1:]
            t = t.get_root()

        self.set_status("Snoozing %s" % (t.full_title()))
        try:
            t.snooze(text)
        except ValueError, e:
            self.set_status(str(e))
            return

        self._save_task(t, "Snoozed %s" % (t.full_title()))

    def _activate_vicommand(self, text):
        """make a few vi commands work, like :w and :q and :e"""
        if text == "w":
            self._save_task_cb()
        elif text == "q":
            self._quit()
        #elif text == "wq": # FIXME: should save all files instead of just one?
        #    self._save_task_cb()
        #    self._quit()
        elif text == "e":
            self._reload_all()
        # allow the user to prepend ":" to some other commands
        elif text.startswith('/'):
            self._activate_search(text[1:])
        elif text.startswith('z'):
            self._activate_snooze(text[1:])
        else:
            self.set_status("Unknown vi command: %s" % (text))

    def _quit(self, *args):
        self.gtk.main_quit()

def set_defaults(cfg):
    cfg.Taskfile = Taskfile

    # FIXME: remove this config option, since I can't make the regexes update 
    # with it
    cfg.doc(datefmt="How to format dates.")
    cfg.default(datefmt="%Y-%m-%d@%H:%M")

    cfg.doc(files="A list of Taskfile objects.")
    cfg.default(files=[])

    cfg.doc(hide_zero="Should we hide zero-score items?")
    cfg.default(hide_zero=True)

    cfg.doc(hide_done="Should we hide completed tasks?")
    cfg.default(hide_done=True)

    return cfg

def create_config(cfg):
    import os
    dirname = "%s/.%s" % (os.environ["HOME"], program_name)
    rcpath = "%s/rc" % (dirname)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    if not os.path.exists(rcpath):
        open(rcpath, "w")
        cfg.save()

def main(args):
    import pycfg
    global cfg
    cfg = pycfg.config(program_name)
    set_defaults(cfg)
    create_config(cfg)
    # FIXME: tends to clobber the config data if loading had any errors
    # (which means losing all the config when it saves later)
    cfg.load()

    # TODO: handle command line
    # TODO: export plain text summary
    # TODO: config option for default mode (gui, etc)
    # TODO: config option for save-and-exit immediately

    #ml = Masterlist(cfg.files)
    ml = MasterlistGUI(cfg.files)
    ml.load()
    ml.sort()
    #ml.print_list()
    ml.run()

    #for tf in ml.taskfiles:
    #    print repr(tf)
    #    print tf.to_otl()

    # FIXME: only save if modified
    # (or only save if the user says to do so, in GUI?)
    #cfg.save()

if __name__ == "__main__":
    import sys
    main(sys.argv[1:])
